@@ -34,6 +34,8 @@
 
 #include <glib/gi18n.h>
 #include <gtk/gtkx.h>
+#include <libaccounts-glib/ag-account-service.h>
+#include <libaccounts-glib/ag-auth-data.h>
 #include <libaccounts-glib/ag-manager.h>
 #include <libaccounts-glib/ag-provider.h>
 #include <libsignon-glib/signon-auth-session.h>
@@ -53,12 +55,15 @@ struct _ApOAuthPluginPrivate
     GHashTable *oauth_params;
     GHashTable *account_oauth_params;
     GtkSocket *socket;
+    AgAuthData *auth_data;
     SignonIdentity *identity;
     SignonAuthSession *auth_session;
     GDBusConnection *connection;
     GCancellable *cancellable;
     gboolean identity_stored;
     gboolean deleting_identity;
+    gboolean authenticating;
+    gboolean storing_account;
 };
 
 static const gchar signon_id[] = AP_PLUGIN_CREDENTIALS_ID_FIELD;
@@ -76,6 +81,51 @@ G_DEFINE_TYPE (ApOAuthPlugin, ap_oauth_plugin, AP_TYPE_PLUGIN);
 
 #define AP_OAUTH_PLUGIN_PRIV(obj) (AP_OAUTH_PLUGIN(obj)->priv)
 
+static const gchar *
+get_mechanism(ApOAuthPluginPrivate *priv)
+{
+    const gchar *mechanism = NULL;
+
+    if (priv->mechanism != NULL)
+        return priv->mechanism;
+
+    if (priv->auth_data != NULL)
+        mechanism = ag_auth_data_get_mechanism (priv->auth_data);
+
+    if (mechanism == NULL)
+        mechanism = oauth_mechanisms[AP_OAUTH_MECHANISM_USER_AGENT];
+
+    return mechanism;
+}
+
+static GVariant *
+value_to_variant (GValue *value)
+{
+    const GVariantType *type;
+
+    switch (G_VALUE_TYPE (value))
+    {
+    case G_TYPE_STRING: type = G_VARIANT_TYPE_STRING; break;
+    case G_TYPE_BOOLEAN: type = G_VARIANT_TYPE_BOOLEAN; break;
+    case G_TYPE_UCHAR: type = G_VARIANT_TYPE_BYTE; break;
+    case G_TYPE_INT: type = G_VARIANT_TYPE_INT32; break;
+    case G_TYPE_UINT: type = G_VARIANT_TYPE_UINT32; break;
+    case G_TYPE_INT64: type = G_VARIANT_TYPE_INT64; break;
+    case G_TYPE_UINT64: type = G_VARIANT_TYPE_UINT64; break;
+    case G_TYPE_DOUBLE: type = G_VARIANT_TYPE_DOUBLE; break;
+    default:
+        if (G_VALUE_TYPE (value) == G_TYPE_STRV)
+        {
+            type = G_VARIANT_TYPE_STRING_ARRAY;
+            break;
+        }
+        g_warning ("Unsupported type %s", G_VALUE_TYPE_NAME (value));
+        return NULL;
+    }
+
+    return g_dbus_gvalue_to_gvariant (value, type);
+}
+
 static gboolean
 emit_finished (ApPlugin *plugin)
 {
@@ -88,25 +138,34 @@ finish_if_ready (ApOAuthPlugin *self)
 {
     ApOAuthPluginPrivate *priv = self->priv;
 
-    if (!priv->deleting_identity)
-    {
-        g_idle_add ((GSourceFunc)emit_finished, self);
-    }
+    /* Check if some asynchronous operations are still running */
+    if (priv->authenticating) return;
+    if (priv->storing_account) return;
+    if (priv->deleting_identity) return;
+
+    g_idle_add ((GSourceFunc)emit_finished, self);
 }
 
 static void
 identity_removed_cb (SignonIdentity *identity, const GError *error,
                      gpointer user_data)
 {
-    ApOAuthPlugin *self = AP_OAUTH_PLUGIN (user_data);
+    ApOAuthPlugin *self;
 
     if (G_UNLIKELY (error != NULL))
     {
+        if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+        {
+            return;
+        }
         g_critical ("Error removing identity: %s", error->message);
         /* No special handling, we are quitting anyways */
     }
 
+    self = AP_OAUTH_PLUGIN (user_data);
+
     self->priv->deleting_identity = FALSE;
+    self->priv->identity_stored = FALSE;
     finish_if_ready (self);
 }
 
@@ -121,7 +180,8 @@ finish_with_cleanup (ApOAuthPlugin *self)
     }
 
     if (priv->identity != NULL &&
-        priv->identity_stored)
+        priv->identity_stored &&
+        !priv->deleting_identity)
     {
         priv->deleting_identity = TRUE;
         signon_identity_remove (priv->identity, identity_removed_cb, self);
@@ -145,14 +205,27 @@ finish_with_cancellation (ApOAuthPlugin *self)
 }
 
 static void
-account_store_cb (AgAccount *account, const GError *error, gpointer user_data)
+account_store_cb (GObject *source_object, GAsyncResult *res,
+                  gpointer user_data)
 {
-    ApOAuthPlugin *self = AP_OAUTH_PLUGIN (user_data);
+    ApOAuthPlugin *self;
+    GError *error = NULL;
+
+    ag_account_store_finish (AG_ACCOUNT (source_object), res, &error);
+    if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+    {
+        g_error_free (error);
+        return;
+    }
+
+    self = AP_OAUTH_PLUGIN (user_data);
+    self->priv->storing_account = FALSE;
 
     if (G_UNLIKELY (error != NULL))
     {
         g_critical ("Account write error: %s", error->message);
         finish_with_error (self, error);
+        g_error_free (error);
         return;
     }
 
@@ -170,15 +243,15 @@ unset_authentication_settings (AgAccount *account,
 {
     AgAccountSettingIter iter;
     const gchar *key;
-    const GValue *value;
+    GVariant *variant;
     gchar *prefix, *full_key;
 
     prefix = g_strdup_printf ("auth/%s/%s/", method, mechanism);
     ag_account_settings_iter_init (account, &iter, prefix);
-    while (ag_account_settings_iter_next (&iter, &key, &value))
+    while (ag_account_settings_iter_get_next (&iter, &key, &variant))
     {
         full_key = g_strconcat (prefix, key, NULL);
-        ag_account_set_value (account, full_key, NULL);
+        ag_account_set_variant (account, full_key, NULL);
         g_free (full_key);
     }
     g_free (prefix);
@@ -190,10 +263,8 @@ store_authentication_parameters (ApOAuthPlugin *self, AgAccount *account)
     ApOAuthPluginPrivate *priv = self->priv;
     GHashTableIter iter;
     gpointer key, value;
-    GValue v_method = G_VALUE_INIT;
-    GValue v_mechanism = G_VALUE_INIT;
-    GHashTable *parameters;
     gchar *full_key;
+    const gchar *mechanism;
     const gchar *username;
 
     username = ap_plugin_get_username ((ApPlugin *)self);
@@ -202,32 +273,28 @@ store_authentication_parameters (ApOAuthPlugin *self, AgAccount *account)
         ag_account_set_display_name (account, username);
     }
 
-    g_value_init (&v_method, G_TYPE_STRING);
-    g_value_set_static_string (&v_method, oauth_method);
-    ag_account_set_value (account, "auth/method", &v_method);
-
-    g_value_init (&v_mechanism, G_TYPE_STRING);
-    g_value_set_static_string (&v_mechanism, priv->mechanism);
-    ag_account_set_value (account, "auth/mechanism", &v_mechanism);
-
+    mechanism = get_mechanism (priv);
     /* Delete any existing parameter (if the account is not new) */
     if (account->id != 0)
     {
         unset_authentication_settings (account,
-                                       oauth_method, priv->mechanism);
+                                       oauth_method, mechanism);
     }
 
     /* Add all the provider-specific OAuth parameters. */
-    parameters = priv->account_oauth_params != NULL ?
-        priv->account_oauth_params : priv->oauth_params;
-    g_hash_table_iter_init (&iter, parameters);
-    while (g_hash_table_iter_next (&iter, &key, &value))
+    if (priv->account_oauth_params != NULL)
     {
-        full_key = g_strdup_printf ("auth/%s/%s/%s",
-                                    oauth_method, priv->mechanism,
-                                    (gchar *)key);
-        ag_account_set_value (account, full_key, value);
-        g_free (full_key);
+        g_hash_table_iter_init (&iter, priv->account_oauth_params);
+        while (g_hash_table_iter_next (&iter, &key, &value))
+        {
+            GVariant *variant = value_to_variant (value);
+            full_key = g_strdup_printf ("auth/%s/%s/%s",
+                                        oauth_method, mechanism,
+                                        (gchar *)key);
+            ag_account_set_variant (account, full_key, variant);
+            g_variant_unref (variant);
+            g_free (full_key);
+        }
     }
 }
 
@@ -239,7 +306,9 @@ store_account (ApOAuthPlugin *self)
     account = ap_plugin_get_account ((ApPlugin *)self);
     store_authentication_parameters (self, account);
 
-    ag_account_store (account, account_store_cb, self);
+    self->priv->storing_account = TRUE;
+    ag_account_store_async (account, self->priv->cancellable,
+                            account_store_cb, self);
 }
 
 static void
@@ -277,12 +346,33 @@ query_username (ApOAuthPlugin *self)
 }
 
 static void
-auth_session_process_cb (SignonAuthSession *auth_session,
-                         GHashTable *session_data,
-                         const GError *error, gpointer user_data)
+auth_session_process_cb (GObject *source_object,
+                         GAsyncResult *res,
+                         gpointer user_data)
 {
-    ApOAuthPlugin *self = AP_OAUTH_PLUGIN (user_data);
+    ApOAuthPlugin *self;
+    SignonAuthSession *auth_session = SIGNON_AUTH_SESSION (source_object);
     AgAccount *account;
+    GVariant *session_data;
+    GError *error = NULL;
+
+    session_data = signon_auth_session_process_finish (auth_session,
+                                                       res,
+                                                       &error);
+    if (session_data != NULL)
+    {
+        /* We don't care about the response */
+        g_variant_unref (session_data);
+    }
+
+    if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED))
+    {
+        g_error_free (error);
+        return;
+    }
+
+    self = AP_OAUTH_PLUGIN (user_data);
+    self->priv->authenticating = FALSE;
 
     if (G_UNLIKELY (error != NULL))
     {
@@ -292,11 +382,13 @@ auth_session_process_cb (SignonAuthSession *auth_session,
             /* Don't report an error if the session has been cancelled by the
              * user */
             finish_with_cancellation (self);
-            return;
         }
-
-        g_warning ("AuthSession error: %s", error->message);
-        finish_with_error (self, error);
+        else
+        {
+            g_warning ("AuthSession error: %s", error->message);
+            finish_with_error (self, error);
+        }
+        g_error_free (error);
         return;
     }
 
@@ -337,46 +429,70 @@ cookies_to_variant (GHashTable *cookies)
     return g_variant_builder_end (&builder);
 }
 
-static void
-start_authentication_process (ApOAuthPlugin *self)
+static GVariant *
+prepare_session_data (ApOAuthPlugin *self)
 {
     ApOAuthPluginPrivate *priv = self->priv;
-    GHashTable *session_data;
+    GVariant *session_data;
+    GVariantBuilder builder;
     GHashTableIter iter;
-    GValue v_window_id = G_VALUE_INIT;
-    GValue v_embedded = G_VALUE_INIT;
     GHashTable *cookies;
-    GValue v_cookies = G_VALUE_INIT;
     gpointer key, value;
-    GError *error = NULL;
 
-    session_data = g_hash_table_new (g_str_hash, g_str_equal);
+    g_variant_builder_init (&builder, G_VARIANT_TYPE_VARDICT);
+    if (priv->socket != NULL)
+    {
+        guint32 window_id = gtk_socket_get_id (priv->socket);
 
-    g_value_init (&v_window_id, G_TYPE_UINT);
-    g_value_set_uint (&v_window_id, gtk_socket_get_id (priv->socket));
-    g_hash_table_insert (session_data, "WindowId", &v_window_id);
-
-    g_value_init (&v_embedded, G_TYPE_BOOLEAN);
-    g_value_set_boolean (&v_embedded, TRUE);
-    g_hash_table_insert (session_data, "Embedded", &v_embedded);
+        g_variant_builder_add (&builder, "{sv}",
+                               "WindowId",
+                               g_variant_new_uint32 (window_id));
+        g_variant_builder_add (&builder, "{sv}",
+                               "Embedded",
+                               g_variant_new_boolean (TRUE));
+    }
 
     cookies = ap_plugin_get_cookies ((ApPlugin *)self);
     if (cookies != NULL &&
         !ap_plugin_get_ignore_cookies ((ApPlugin *)self))
     {
-        g_value_init (&v_cookies, G_TYPE_VARIANT);
-        g_value_set_variant (&v_cookies, cookies_to_variant (cookies));
-        g_hash_table_insert (session_data, "Cookies", &v_cookies);
+        g_variant_builder_add (&builder, "{sv}",
+                               "Cookies",
+                               cookies_to_variant (cookies));
     }
 
     /* Add all the provider-specific OAuth parameters. */
-    g_hash_table_iter_init (&iter, priv->oauth_params);
-    while (g_hash_table_iter_next (&iter, &key, &value))
+    if (priv->oauth_params != NULL)
     {
-        g_hash_table_insert (session_data, key, value);
+        g_hash_table_iter_init (&iter, priv->oauth_params);
+        while (g_hash_table_iter_next (&iter, &key, &value))
+        {
+            GVariant *variant = value_to_variant (value);
+            g_variant_builder_add (&builder, "{sv}", key, variant);
+            g_variant_unref (variant);
+        }
     }
 
-    /* Finally, get the authentication started */
+    /* Merge the session parameters built so far with the provider's default
+     * parameters */
+    session_data = g_variant_builder_end (&builder);
+    if (priv->auth_data != NULL)
+    {
+        session_data = ag_auth_data_get_login_parameters (priv->auth_data,
+                                                          session_data);
+    }
+    return session_data;
+}
+
+static void
+start_authentication_process (ApOAuthPlugin *self)
+{
+    ApOAuthPluginPrivate *priv = self->priv;
+    GVariant *session_data;
+    GError *error = NULL;
+
+    session_data = prepare_session_data (self);
+
     priv->auth_session = signon_identity_create_session (priv->identity,
                                                          oauth_method,
                                                          &error);
@@ -387,11 +503,12 @@ start_authentication_process (ApOAuthPlugin *self)
         g_clear_error (&error);
         return;
     }
-    signon_auth_session_process (priv->auth_session, session_data,
-                                 priv->mechanism,
-                                 auth_session_process_cb, self);
-    g_hash_table_unref (session_data);
-    g_value_unset (&v_cookies);
+    priv->authenticating = TRUE;
+    signon_auth_session_process_async (priv->auth_session, session_data,
+                                       get_mechanism (priv),
+                                       priv->cancellable,
+                                       auth_session_process_cb,
+                                       self);
 }
 
 static void
@@ -401,7 +518,7 @@ identity_store_cb (SignonIdentity *identity, guint32 id,
     ApOAuthPlugin *self = AP_OAUTH_PLUGIN (user_data);
     ApOAuthPluginPrivate *priv = self->priv;
     AgAccount *account;
-    GValue v_id = G_VALUE_INIT;
+    GVariant *v_id;
 
     if (G_UNLIKELY (error != NULL))
     {
@@ -413,10 +530,9 @@ identity_store_cb (SignonIdentity *identity, guint32 id,
     priv->identity_stored = TRUE;
 
     /* store the identity ID into the account settings */
-    g_value_init (&v_id, G_TYPE_UINT);
-    g_value_set_uint (&v_id, id);
+    v_id = g_variant_new_uint32 (id);
     account = ap_plugin_get_account ((ApPlugin *)self);
-    ag_account_set_value (account, signon_id, &v_id);
+    ag_account_set_variant (account, signon_id, v_id);
 
     if (priv->socket != NULL)
     {
@@ -438,12 +554,6 @@ setup_authentication (ApOAuthPlugin *self)
     const gchar *username;
     const gchar *secret;
 
-    if (G_UNLIKELY (self->priv->oauth_params == NULL))
-    {
-        g_critical ("OAuth parameters missing!");
-        return;
-    }
-
     username = ap_plugin_get_username ((ApPlugin *)self);
     secret = ap_plugin_get_password ((ApPlugin *)self);
     if (secret == NULL) secret = "";
@@ -491,7 +601,7 @@ reauthenticate_account_cb (GDBusConnection *connection,
     }
 
     g_assert (result != NULL);
-    g_variant_get (result, G_VARIANT_TYPE ("(b)"), &authenticated);
+    g_variant_get (result, "(b)", &authenticated);
     ap_plugin_set_need_authentication ((ApPlugin *)self, !authenticated);
     finish_with_cleanup (self);
 }
@@ -666,10 +776,28 @@ ap_oauth_plugin_init (ApOAuthPlugin *self)
 {
     self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self, AP_TYPE_OAUTH_PLUGIN,
                                               ApOAuthPluginPrivate);
-    self->priv->mechanism = oauth_mechanisms[AP_OAUTH_MECHANISM_USER_AGENT];
+    self->priv->mechanism = NULL;
     self->priv->cancellable = g_cancellable_new ();
 }
 
+static void
+ap_oauth_plugin_constructed (GObject *object)
+{
+    ApOAuthPluginPrivate *priv = AP_OAUTH_PLUGIN_PRIV (object);
+    AgAccount *account;
+    AgAccountService *account_service;
+
+    G_OBJECT_CLASS (ap_oauth_plugin_parent_class)->constructed (object);
+
+    account = ap_plugin_get_account (AP_PLUGIN (object));
+    if (account != NULL)
+    {
+        account_service = ag_account_service_new (account, NULL);
+        priv->auth_data = ag_account_service_get_auth_data (account_service);
+        g_object_unref (account_service);
+    }
+}
+
 static void
 ap_oauth_plugin_set_property (GObject *object, guint property_id,
                         const GValue *value, GParamSpec *pspec)
@@ -722,6 +850,12 @@ ap_oauth_plugin_dispose (GObject *object)
         priv->oauth_params = NULL;
     }
 
+    if (priv->auth_data)
+    {
+        ag_auth_data_unref (priv->auth_data);
+        priv->auth_data = NULL;
+    }
+
     if (priv->identity)
     {
         g_object_unref (priv->identity);
@@ -834,6 +968,7 @@ ap_oauth_plugin_class_init (ApOAuthPluginClass *klass)
 
     g_type_class_add_private (object_class, sizeof (ApOAuthPluginPrivate));
 
+    object_class->constructed = ap_oauth_plugin_constructed;
     object_class->set_property = ap_oauth_plugin_set_property;
     object_class->get_property = ap_oauth_plugin_get_property;
     object_class->dispose = ap_oauth_plugin_dispose;
@@ -862,7 +997,10 @@ ap_oauth_plugin_class_init (ApOAuthPluginClass *klass)
  * @oauth_params: (element-type utf8 GValue): a dictionary of OAuth parameters.
  *
  * Sets the dictionary of OAuth parameters to be used when authenticating the
- * account.
+ * account. Note that these parameters are not stored in the account and will
+ * be used only by the account plugin.
+ * To set the authentication parameters used by applications, use
+ * ap_oauth_plugin_set_account_oauth_parameters().
  */
 void
 ap_oauth_plugin_set_oauth_parameters (ApOAuthPlugin *self,
@@ -883,9 +1021,7 @@ ap_oauth_plugin_set_oauth_parameters (ApOAuthPlugin *self,
  * Sets the dictionary of OAuth parameters to be used by client applications
  * when authenticating the account. These are the parameters which will be
  * stored into the account configuration (those used by the plugin itself when
- * authenticating are those set with ap_oauth_plugin_set_oauth_parameters());
- * if this method is not called, then the parameters set with
- * ap_oauth_plugin_set_oauth_parameters() will be stored into the account.
+ * authenticating are those set with ap_oauth_plugin_set_oauth_parameters()).
  */
 void
 ap_oauth_plugin_set_account_oauth_parameters (ApOAuthPlugin *self,
@@ -912,3 +1048,11 @@ void ap_oauth_plugin_set_mechanism (ApOAuthPlugin *self,
     g_return_if_fail (mechanism < AP_OAUTH_MECHANISM_LAST);
     self->priv->mechanism = oauth_mechanisms[mechanism];
 }
+
+#ifdef BUILDING_UNIT_TESTS
+GVariant *prepare_session_data_test (ApOAuthPlugin *self);
+GVariant *prepare_session_data_test (ApOAuthPlugin *self)
+{
+    return prepare_session_data (self);
+}
+#endif
